#undef DEBUG_DMA

#include <stdio.h>
#include <math.h>
#include <signal.h>
#include "dmavif1.h"
#include "dmagif.h"
#include "screen.h" 
#include "geommath.h"
#include "camera.h"
#include "p2lassert.h"

int loop_again=1;

// this is executed when ctrl-c is pressed, ideally it stops the loop.
// the deinit functions have to be at the end of the main after the loop.
void sig_handle(int sig)
{
	loop_again=0;
}


// distribute sps2 specific global pointers to relevant classes
void  DistribGlobPtrs(void * pDescCached, void * pDescUncached)
{
	CDmaWrap::m_pDescCached = (sps2Memory_t *) pDescCached;
	CDmaWrap::m_pDescUncached = (sps2Memory_t *) pDescUncached;
}


extern unsigned char uTexture[];
extern unsigned char uBumpPos[];
extern unsigned char uBumpNeg[];


void MakeBiNormals();
void DrawTriangles(const Mat44 &mLocToWorld, const Mat44 &mLocToBitmap, float fDir, bool bABE);
void DrawTriangles(const Mat44 &mLocToBitmap);
void UploadTexture(void * pTex, uint32 uFirstPage, uint64 uClamp, uint64 uAlpha, uint64 uColClamp);

// PS2 Camera setup
// BMPSize(w, -h, 2^24 - 1)
Vec4 vCamWorldPos;
void * pAllDmaMem;
uint32 * PaletteUnc, * PaletteCach;

void DownLoadPals();
CDmaVif1 BuildAddRGBintoAlpha(uint32 uDmaMem, uint32 lFrame,uint32 lTemp);

int iSPS2Device;
CScreen *cScreen=NULL;

void delete_screen()
{
	if (cScreen!=NULL) delete cScreen;
	cScreen=NULL;
}

int main(int argc, char **argv)
{

	// Open sps2-library and PS2Linux stuff
	iSPS2Device= sps2Init();

	if (iSPS2Device<0) {
	  printf("Can't open SPS2 device.  Is the module loaded?\n");
	  exit(-1);
	}

	uint32 uAllocDmaMegs = 4;
	sps2Memory_t * pDescCached = sps2Allocate(uAllocDmaMegs*1024*1024, SPS2_MAP_BLOCK_4K | SPS2_MAP_CACHED, iSPS2Device);
	sps2Memory_t * pDescUncached = sps2Remap(pDescCached, SPS2_MAP_UNCACHED, iSPS2Device);
	DistribGlobPtrs(pDescCached, pDescUncached);
	pAllDmaMem = pDescCached->pvStart;	
	for (int sig=0; sig<128; sig++)	// catch as many signals as possible
		signal(sig, sig_handle);

	
	// Setup Screen
	cScreen=new CScreen;
	atexit(delete_screen);
	cScreen->ScreenSetup(pAllDmaMem, 16);
	uint32 uFirstGSPage = cScreen->GetFirstFreeGsPage();

	// Setup Camera
	float fScrWdth = cScreen->GetWidth();
	float fScrHgt = cScreen->GetHeight();
	float fDepthZ = cScreen->GetDepth();
	CCamera cCam;
	cCam.SetBitmapSize(fScrWdth, -fScrHgt, fDepthZ);
	cCam.SetViewPortOffset(2048, 2048, fDepthZ);
	cCam.SetClipPlanes(-10, -10000);
	cCam.SetEyePos(Vec3(0,0,1));
	float fFov = 0.524f;	// total view of 60 degrees
	float fMinU = -tan(fFov);
	float fMinV = (fMinU*3)/4;
	cCam.SetBoundVRC(fMinU, fMinV, -fMinU, -fMinV);
	cCam.SetPosViewPlane(Vec3(-400,400,400));
	cCam.SetRotViewPlane(Vec3(-400,400,400), Vec3(0,1,0));
	//cCam.SetPosViewPlane(Vec3(0,0,800));
	//cCam.SetRotViewPlane(Vec3(0,0,1), Vec3(0,1,0));

	vCamWorldPos = cCam.GetTransformCamToWorld()*Vec4(0,0,1,1);

	MakeBiNormals();

	// Move textures to DmaMem
	void * pTexCol = (((char *) pAllDmaMem) + 4096*16);
	void * pTexBumpPos = (((char *) pTexCol) + 256*256*4);
	void * pTexBumpNeg = (((char *) pTexBumpPos) + 256*256*4);
	memcpy(pTexCol, uTexture, 256*256*4);
	memcpy(pTexBumpPos, uBumpPos, 256*256*4);
	memcpy(pTexBumpNeg, uBumpNeg, 256*256*4);

	PaletteUnc = (uint32 *) (((char *) pDescUncached->pvStart) + 4096*16 + 256*256*4*3);
	PaletteCach = (uint32 *) (((char *) pTexBumpNeg) + 256*256*4);	

	// Prebuild chains	
	uint32 uDraw1 = cScreen->GetDrawBuff1();
	uint32 uDraw2 = cScreen->GetDrawBuff2();
	uint32 uDmaMem1 = (uint32) (((char *) pAllDmaMem) + 1024*1024*2);
	uint32 uDmaMem2 = (uint32) (((char *) pAllDmaMem) + 1024*1024*3);	
	CDmaVif1 cVif1Add1 = BuildAddRGBintoAlpha(uDmaMem1, uDraw1, uFirstGSPage);
	CDmaVif1 cVif1Add2 = BuildAddRGBintoAlpha(uDmaMem2, uDraw2, uFirstGSPage);

	// Flush Cache
	sps2FlushCache(iSPS2Device);

	while(loop_again)
	{
		// Setup transformation
		static float fAx=0, fAy=0, fAz=0;
		const float delta_rot = 0.25f * 3.14159265f / 180.0f;
		fAx += delta_rot     ; if(fAx > 3.14159265f) fAx -= 2.0f * 3.14159265f;
		fAy += delta_rot/2.0f; if(fAy > 3.14159265f) fAy -= 2.0f * 3.14159265f;
		fAz += delta_rot/3.0f; if(fAz > 3.14159265f) fAz -= 2.0f * 3.14159265f;

		Mat44 mLocToWorld;
		LoadIdentity(&mLocToWorld);
		LoadRotation(&mLocToWorld, fAx, fAy, fAz);

		Mat44 mLocToBitmap = cCam.GetWorldToBitmapTransform() * mLocToWorld;

		// posTex*posLgt
		UploadTexture(pTexBumpPos, uFirstGSPage, CLAMP_SET(0, 0, 0,0,0,0), ALPHA_SET(2, 2, 2, 0, 128), COLCLAMP_SET(1));
		DrawTriangles(mLocToWorld, mLocToBitmap, 1, false);
		//GsSyncPath(0,0);
		
		// negTex*negLgt 
		UploadTexture(pTexBumpNeg, uFirstGSPage, CLAMP_SET(0, 0, 0,0,0,0), ALPHA_SET(0, 2, 2, 1, 128), 0xffffffff);
		DrawTriangles(mLocToWorld, mLocToBitmap, -1, true);
		//GsSyncPath(0,0);


		// negTex*posLgt 
		UploadTexture(pTexBumpNeg, uFirstGSPage, 0xffffffff, ALPHA_SET(1, 0, 2, 2, 128), COLCLAMP_SET(0));
		DrawTriangles(mLocToWorld, mLocToBitmap, 1, true);
		//GsSyncPath(0,0);

		// posTex*negLgt 
		UploadTexture(pTexBumpPos, uFirstGSPage, CLAMP_SET(0, 0, 0,0,0,0), 0xffffffff, 0xffffffff);
		DrawTriangles(mLocToWorld, mLocToBitmap, -1, true);
		//GsSyncPath(0,0);

		DownLoadPals();
		if(cScreen->GetDrawBuff1() == cScreen->GetDrawFptr())
		{
			cVif1Add1.Launch(iSPS2Device);
			cVif1Add1.Wait();
		}
		else
		{
			cVif1Add2.Launch(iSPS2Device);
			cVif1Add2.Wait();
		}
		//GsSyncPath(0,0);

		CDmaVif1 dmaMiniVif1(pAllDmaMem, 6);
		dmaMiniVif1.AttachSrcEndTag(DIRECT(5)<<32);
		dmaMiniVif1.add64(GIFTAG(4, 1, 0, 0, 0, 1));
		dmaMiniVif1.add64(GIFTAG_REGS1(AD));
		dmaMiniVif1.add64(ZBUF_SET(cScreen->GetZptr(), 1, 0));
		dmaMiniVif1.add64(ZBUF_1);
		dmaMiniVif1.add64(TEST_SET(0, 0, 0, 0, 0, 0, 1, 2));
		dmaMiniVif1.add64(TEST_1);
		dmaMiniVif1.add64(FRAME_SET(cScreen->GetDrawFptr(),cScreen->GetWidth()/64,0,0));
		dmaMiniVif1.add64(FRAME_1);

		uint32 uOffsX = (2048-(cScreen->GetWidth()>>1))*16;				// Now center in render will always be 2048, 2048
		uint32 uOffsY = (2048-(cScreen->GetHeight()>>1))*16;
		dmaMiniVif1.add64(XYOFFSET_SET(uOffsX, uOffsY));		// XYOffs = (0x7000, 0x7200)
		dmaMiniVif1.add64(XYOFFSET_1);
		dmaMiniVif1.EndPacket();

		dmaMiniVif1.Launch(iSPS2Device);
		dmaMiniVif1.Wait();
		//GsSyncPath(0,0);
		
		UploadTexture(pTexCol, uFirstGSPage, CLAMP_SET(0, 0, 0,0,0,0), ALPHA_SET(0,2,1,2,128), COLCLAMP_SET(1));
		DrawTriangles(mLocToBitmap);


		cScreen->Swap();
	}

	return 0;
}

void UploadTexture(void * pTex, uint32 uFirstPage, uint64 uClamp, uint64 uAlpha, uint64 uColClamp)
{
	CDmaVif1 dmaVif1(pAllDmaMem, 4096);

	uint32 uDstPtr = (uFirstPage*2048)/64;//0x1a40;
	uint32 uQSize = (256*256*4)/16;

	dmaVif1.AttachSrcCntTag(DIRECT(6)<<32);
	dmaVif1.add64(GIFTAG(4, 0, 0, 0, 0, 1));	// GIFTAG(NLOOP, EOP, PRE, PRIM, FLG, NREG)
	dmaVif1.add64(GIFTAG_REGS1(AD));
	dmaVif1.add64(BITBLTBUF_SET(0, 0, 0, uDstPtr, 4, 0));		// 0x41a40 00000000L
	dmaVif1.add64(BITBLTBUF);
	dmaVif1.add64(TRXPOS_SET(0, 0, 0, 0, 0));
	dmaVif1.add64(TRXPOS);
	dmaVif1.add64(TRXREG_SET(256, 256));
	dmaVif1.add64(TRXREG);
	dmaVif1.add64(TRXDIR_SET(0));
	dmaVif1.add64(TRXDIR);
	dmaVif1.add64(GIFTAG(uQSize, 0, 0, 0, 2, 0));	// GIFTAG(NLOOP, EOP, PRE, PRIM, FLG, NREG)
	dmaVif1.add64(0);
	dmaVif1.EndPacket();

	for(int page=0; page<((256*256*4)/4096); page++)
		dmaVif1.AttachSrcRefTag(((char *) pTex)+page*4096, 4096/16, DIRECT((4096/16))<<32);

	uint64 uNloop = 11;
	if(uAlpha==0xffffffff) --uNloop;
	if(uClamp==0xffffffff) --uNloop;
	if(uColClamp==0xffffffff) --uNloop;

	dmaVif1.AttachSrcEndTag(DIRECT(uNloop+1)<<32);
	dmaVif1.add64(GIFTAG(uNloop, 1, 0, 0, 0, 1));	// GIFTAG(NLOOP, EOP, PRE, PRIM, FLG, NREG)
	dmaVif1.add64(GIFTAG_REGS1(AD));
	dmaVif1.add64(TEST_SET(0, 0, 0, 0, 0, 0, 1, 2));	// 0x70000
	dmaVif1.add64(TEST_1);
	if(uAlpha!=0xffffffff)
	{
		dmaVif1.add64(uAlpha);	// 0x44
		dmaVif1.add64(ALPHA_1);
	}
	dmaVif1.add64(PABE_SET(0));
	dmaVif1.add64(PABE);
	dmaVif1.add64(TEXA_SET(0x7f, 1, 0x81));	//     0x810000807fL
	dmaVif1.add64(TEXA);
	dmaVif1.add64(FBA_SET(0));
	dmaVif1.add64(FBA_1);
	dmaVif1.add64(0);		// NOP data always 0
	dmaVif1.add64(0xf);	// NOP instruction
	dmaVif1.add64(TEX1_SET(0, 0, 3, 0, 0, 0, 0));		// 0x60
	dmaVif1.add64(TEX1_1);
	dmaVif1.add64(TEX0_SET(uDstPtr, 4, 0, 8, 8, 1, 0, 0, 0, 0, 0, 0));	// 0x620011a40L
	dmaVif1.add64(TEX0_1);
	if(uClamp!=0xffffffff)
	{
		dmaVif1.add64(uClamp);
		dmaVif1.add64(CLAMP_1);
	}
	if(uColClamp!=0xffffffff)
	{
		dmaVif1.add64(uColClamp);
		dmaVif1.add64(COLCLAMP);
	}
	dmaVif1.add64(0);
	dmaVif1.add64(TEXFLUSH);
	dmaVif1.EndPacket();
	dmaVif1.Launch(iSPS2Device);
	dmaVif1.Wait();
}

const float fSize = 200;

uint32 uIndices[] = {
						0,1,2,0,2,3,
						4,5,6,4,6,7,
						8,9,10,8,10,11,
						12,13,14,12,14,15,
						16,17,18,16,18,19,
						20,21,22,20,22,23
					};

Vec3 Points[] =	{
					// +Z axis
					Vec3(-fSize, -fSize, fSize), Vec3(fSize, -fSize, fSize), Vec3(fSize, fSize, fSize), Vec3(-fSize, fSize, fSize),

					// +X axis
					Vec3(fSize, -fSize, fSize), Vec3(fSize, -fSize, -fSize), Vec3(fSize, fSize, -fSize), Vec3(fSize, fSize, fSize),

					// -Z axis
					Vec3(fSize, -fSize, -fSize), Vec3(-fSize, -fSize, -fSize), Vec3(-fSize, fSize, -fSize), Vec3(fSize, fSize, -fSize),

					// -X axis
					Vec3(-fSize, -fSize, -fSize), Vec3(-fSize, -fSize, fSize), Vec3(-fSize, fSize, fSize), Vec3(-fSize, fSize, -fSize),

					// +Y axis
					Vec3(-fSize, fSize, fSize), Vec3(fSize, fSize, fSize), Vec3(fSize, fSize, -fSize), Vec3(-fSize, fSize, -fSize),

					// -Y axis
					Vec3(-fSize, -fSize, -fSize), Vec3(fSize, -fSize, -fSize), Vec3(fSize, -fSize, fSize), Vec3(-fSize, -fSize, fSize)
				};

Vec3 Normals[] ={
					// +Z axis
					Vec3(0, 0, 1), Vec3(0, 0, 1), Vec3(0, 0, 1), Vec3(0, 0, 1),

					// +X axis
					Vec3(1, 0, 0), Vec3(1, 0, 0), Vec3(1, 0, 0), Vec3(1, 0, 0),

					// -Z axis
					Vec3(0, 0, -1), Vec3(0, 0, -1), Vec3(0, 0, -1), Vec3(0, 0, -1),

					// -X axis
					Vec3(-1, 0, 0), Vec3(-1, 0, 0), Vec3(-1, 0, 0), Vec3(-1, 0, 0),

					// +Y axis
					Vec3(0, 1, 0), Vec3(0, 1, 0), Vec3(0, 1, 0), Vec3(0, 1, 0),

					// -Y axis
					Vec3(0, -1, 0), Vec3(0, -1, 0), Vec3(0, -1, 0), Vec3(0, -1, 0)
				};

Vec3 STQs[] =	{
					Vec3(0,1,1), Vec3(1,1,1), Vec3(1,0,1), Vec3(0,0,1),
					Vec3(0,1,1), Vec3(1,1,1), Vec3(1,0,1), Vec3(0,0,1),
					Vec3(0,1,1), Vec3(1,1,1), Vec3(1,0,1), Vec3(0,0,1),
					Vec3(0,1,1), Vec3(1,1,1), Vec3(1,0,1), Vec3(0,0,1),
					Vec3(0,1,1), Vec3(1,1,1), Vec3(1,0,1), Vec3(0,0,1),
					Vec3(0,1,1), Vec3(1,1,1), Vec3(1,0,1), Vec3(0,0,1),
				};

Vec3 vS[24];
Vec3 vT[24];
Vec3 vBiNorm[24];


void DrawSingleTriangle(const Vec3 &STQ1, const Vec4 &vCol1, const Vec3 &vScr1, const Vec3 &STQ2, const Vec4 &vCol2, const Vec3 &vScr2, const Vec3 &STQ3, const Vec4 &vCol3, const Vec3 &vScr3, bool bABE);


// Render
//Vec3 vLgtDir(0, 1, 0);
//Vec3 vLgtDir(0, 0.707f, 0.707f);
Vec3 vLgtDir(-0.707f, 0.707f, 0);
//Vec3 vLgtDir(-0.707f, 0, 0.707f);
//Vec3 vLgtDir(-0.57735f, 0.57735f, 0.57735f);
//Vec3 vLgtDir(0, 0, 1);

void vClamp(Vec3 * pV)
{
	if(pV->x<0)pV->x=0; if(pV->y<0)pV->y=0; if(pV->z<0)pV->z=0;
	if(pV->x>1)pV->x=1; if(pV->y>1)pV->y=1; if(pV->z>1)pV->z=1;
}

void DrawTriangles(const Mat44 &mLocToWorld, const Mat44 &mLocToBitmap, float fDir, bool bABE)
{
	const uint32 lNumTri = 2*6;
	uint32 iTri;

	Vec4 vT1 = GetColumn(mLocToWorld, 0);
	Vec4 vT2 = GetColumn(mLocToWorld, 1);
	Vec4 vT3 = GetColumn(mLocToWorld, 2);

	Vec3 vX(vT1.x, vT1.y, vT1.z);
	Vec3 vY(vT2.x, vT2.y, vT2.z);
	Vec3 vZ(vT3.x, vT3.y, vT3.z);

	Mat33 mWtoO;
	SetRow(&mWtoO, 0, vX);
	SetRow(&mWtoO, 1, vY);
	SetRow(&mWtoO, 2, vZ);

	Vec3 vLgtObj = mWtoO * vLgtDir;
	Vec4 vCamPosInObj = mWtoO * vCamWorldPos;

	for(iTri=0; iTri<lNumTri; iTri++)
	{
		Vec4 v1=Points[uIndices[3*iTri+0]];
		Vec4 v2=Points[uIndices[3*iTri+1]];
		Vec4 v3=Points[uIndices[3*iTri+2]];
		Vec4 vVecToCam = vCamPosInObj - v1;

		Vec4 d1 = v2 - v1;
		Vec4 d2 = v3 - v1;
		Vec3 vFaceNorm = Cross(Vec3(d1.x, d1.y, d1.z), Vec3(d2.x, d2.y, d2.z));

		// These are big triangles, no need to draw the backside!
		if(vFaceNorm*Vec3(vVecToCam.x, vVecToCam.y, vVecToCam.z)<0)	
			continue;

		Vec4 vSc1 = mLocToBitmap * v1;
		Vec4 vSc2 = mLocToBitmap * v2;
		Vec4 vSc3 = mLocToBitmap * v3;

		Vec3 STQ1=(1.0f/vSc1.w)*STQs[uIndices[3*iTri+0]];
		Vec3 STQ2=(1.0f/vSc2.w)*STQs[uIndices[3*iTri+1]];
		Vec3 STQ3=(1.0f/vSc3.w)*STQs[uIndices[3*iTri+2]];

		Vec3 vScr1 = vSc1;
		Vec3 vScr2 = vSc2;
		Vec3 vScr3 = vSc3;

		// Create Colors
		Vec3 &y1=vBiNorm[uIndices[3*iTri+0]];
		Vec3 &y2=vBiNorm[uIndices[3*iTri+1]];
		Vec3 &y3=vBiNorm[uIndices[3*iTri+2]];

		Vec3 &z1=Normals[uIndices[3*iTri+0]];
		Vec3 &z2=Normals[uIndices[3*iTri+1]];
		Vec3 &z3=Normals[uIndices[3*iTri+2]];

		Vec3 x1 = Normalize(Cross(y1, z1));
		Vec3 x2 = Normalize(Cross(y2, z2));
		Vec3 x3 = Normalize(Cross(y3, z3));

		Mat33 m1, m2, m3;
		SetRow(&m1, 0, x1); SetRow(&m1, 1, y1); SetRow(&m1, 2, z1);
		SetRow(&m2, 0, x2); SetRow(&m2, 1, y2); SetRow(&m2, 2, z2);
		SetRow(&m3, 0, x3); SetRow(&m3, 1, y3); SetRow(&m3, 2, z3);

		Vec3 vLgtTex1 = fDir * (m1 * vLgtObj);
		Vec3 vLgtTex2 = fDir * (m2 * vLgtObj);
		Vec3 vLgtTex3 = fDir * (m3 * vLgtObj);

		vClamp(&vLgtTex1); vClamp(&vLgtTex2); vClamp(&vLgtTex3);

		// Mark Bs uncommented "SelfShading" Hack!!!
		Vec3 vMarkHack(6.0f * (vLgtObj * z1), 6.0f * (vLgtObj * z2), 6.0f * (vLgtObj * z3));
		vClamp(&vMarkHack); vLgtTex1 *= vMarkHack.x; vLgtTex2 *= vMarkHack.y; vLgtTex3 *= vMarkHack.z;
		// Hack End

		Vec4 vCol1(vLgtTex1.x,vLgtTex1.y,vLgtTex1.z,0);
		Vec4 vCol2(vLgtTex2.x,vLgtTex2.y,vLgtTex2.z,0);
		Vec4 vCol3(vLgtTex3.x,vLgtTex3.y,vLgtTex3.z,0);

		DrawSingleTriangle(STQ1, vCol1, vScr1, STQ2, vCol2, vScr2, STQ3, vCol3, vScr3, bABE);
	}
}

void DrawTriangles(const Mat44 &mLocToBitmap)
{
	const uint32 lNumTri = 2*6;
	uint32 iTri;


	for(iTri=0; iTri<lNumTri; iTri++)
	{
		Vec4 v1=Points[uIndices[3*iTri+0]];
		Vec4 v2=Points[uIndices[3*iTri+1]];
		Vec4 v3=Points[uIndices[3*iTri+2]];

		Vec4 vSc1 = mLocToBitmap * v1;
		Vec4 vSc2 = mLocToBitmap * v2;
		Vec4 vSc3 = mLocToBitmap * v3;

		Vec3 STQ1=(1.0f/vSc1.w)*STQs[uIndices[3*iTri+0]];
		Vec3 STQ2=(1.0f/vSc2.w)*STQs[uIndices[3*iTri+1]];
		Vec3 STQ3=(1.0f/vSc3.w)*STQs[uIndices[3*iTri+2]];

		Vec3 vScr1 = vSc1;
		Vec3 vScr2 = vSc2;
		Vec3 vScr3 = vSc3;

	
		Vec4 vCol1(1,1,1,0);
		Vec4 vCol2(1,1,1,0);
		Vec4 vCol3(1,1,1,0);

		DrawSingleTriangle(STQ1, vCol1, vScr1, STQ2, vCol2, vScr2, STQ3, vCol3, vScr3, true);
	}
}

					

void DrawSingleTriangle(const Vec3 &STQ1, const Vec4 &vCol1, const Vec3 &vScr1, const Vec3 &STQ2, const Vec4 &vCol2, const Vec3 &vScr2, const Vec3 &STQ3, const Vec4 &vCol3, const Vec3 &vScr3, bool bABE)
{
	CDmaVif1 dmaVif1(pAllDmaMem, 64);

	uint32 nloop = 9;
	dmaVif1.AttachSrcEndTag(DIRECT(nloop+1)<<32);
	uint32 prim = (bABE) ? PRIM_SET(3, 1, 1, 0, 1, 0, 0, 0, 0) : PRIM_SET(3, 1, 1, 0, 0, 0, 0, 0, 0);
	dmaVif1.add64(GIFTAG(nloop, 1, 1, prim, 0, 1));
	dmaVif1.add64(GIFTAG_REGS1(AD));

	uint32 iZBuf1 = ((int) vScr1.z);
	uint32 iZBuf2 = ((int) vScr2.z);
	uint32 iZBuf3 = ((int) vScr3.z);

	dmaVif1.add64(ST_SET(*((uint32 *) &STQ1.x),*((uint32 *) &STQ1.y)));
	dmaVif1.add64(ST);
	dmaVif1.add64(RGBAQ_SET((int) (vCol1.x*124),(int) (vCol1.y*124),(int) (vCol1.z*124),(int) (vCol1.w*124),*((uint32 *) &STQ1.z)));
	dmaVif1.add64(RGBAQ);
	dmaVif1.add64(XYZ2_SET((int) (vScr1.x*16),(int) (vScr1.y*16),iZBuf1));
	dmaVif1.add64(XYZ2);

	dmaVif1.add64(ST_SET(*((uint32 *) &STQ2.x),*((uint32 *) &STQ2.y)));
	dmaVif1.add64(ST);
	dmaVif1.add64(RGBAQ_SET((int) (vCol2.x*124),(int) (vCol2.y*124),(int) (vCol2.z*124),(int) (vCol2.w*124),*((uint32 *) &STQ2.z)));
	dmaVif1.add64(RGBAQ);
	dmaVif1.add64(XYZ2_SET((int) (vScr2.x*16),(int) (vScr2.y*16),iZBuf2));
	dmaVif1.add64(XYZ2);

	dmaVif1.add64(ST_SET(*((uint32 *) &STQ3.x),*((uint32 *) &STQ3.y)));
	dmaVif1.add64(ST);
	dmaVif1.add64(RGBAQ_SET((int) (vCol3.x*124),(int) (vCol3.y*124),(int) (vCol3.z*124),(int) (vCol3.w*124),*((uint32 *) &STQ3.z)));
	dmaVif1.add64(RGBAQ);
	dmaVif1.add64(XYZ2_SET((int) (vScr3.x*16),(int) (vScr3.y*16),iZBuf3));
	dmaVif1.add64(XYZ2);
	dmaVif1.EndPacket();
	dmaVif1.Launch(iSPS2Device);
	dmaVif1.Wait();

}

#define SMALL_FLOAT 1e-12
/**
 * Taken and modified from shader_dot3_util.cpp from the Dot3Bump example from nVidia
 */
void MakeBiNormals()
{
	uint32 i;
	const uint32 lNumVertices = 24;
	const uint32 lNumTri = 2*6;


	// Clear the basis vectors
	for(i=0; i<lNumVertices; i++)	{
		vS[i]=Vec3(0.0f, 0.0f, 0.0f);
		vT[i]=Vec3(0.0f, 0.0f, 0.0f);
	}

	// Walk through the triangle list and calculate gradiants for each triangle.
	// Sum the results into the S and T components.
    for(i=0; i<lNumTri; i++)
	{	
		Vec3 &vS1=vS[uIndices[3*i+0]];
		Vec3 &vS2=vS[uIndices[3*i+1]];
		Vec3 &vS3=vS[uIndices[3*i+2]];
		Vec3 &vT1=vT[uIndices[3*i+0]];
		Vec3 &vT2=vT[uIndices[3*i+1]];
		Vec3 &vT3=vT[uIndices[3*i+2]];
		Vec3 &v1=Points[uIndices[3*i+0]];
		Vec3 &v2=Points[uIndices[3*i+1]];
		Vec3 &v3=Points[uIndices[3*i+2]];
		Vec3 &Tex1=STQs[uIndices[3*i+0]];
		Vec3 &Tex2=STQs[uIndices[3*i+1]];
		Vec3 &Tex3=STQs[uIndices[3*i+2]];
		

		// x, s, t
		Vec3 edge01, edge02, cp;
		edge01 = Vec3(v2.x-v1.x, Tex2.x-Tex1.x, Tex2.y-Tex1.y);
		edge02 = Vec3(v3.x-v1.x, Tex3.x-Tex1.x, Tex3.y-Tex1.y);
		cp = Cross(edge01, edge02);
		if(fabs(cp.x)>SMALL_FLOAT)
		{
			vS1.x+=-cp.y/cp.x;
			vT1.x+=-cp.z/cp.x;

			vS2.x+=-cp.y/cp.x;
			vT2.x+=-cp.z/cp.x;
			
			vS3.x+=-cp.y/cp.x;
			vT3.x+=-cp.z/cp.x;
		}

		// y, s, t
		edge01 = Vec3(v2.y-v1.y, Tex2.x-Tex1.x, Tex2.y-Tex1.y);
		edge02 = Vec3(v3.y-v1.y, Tex3.x-Tex1.x, Tex3.y-Tex1.y);
		cp = Cross(edge01, edge02);
		if(fabs(cp.x)>SMALL_FLOAT)
		{
			vS1.y+=-cp.y/cp.x;
			vT1.y+=-cp.z/cp.x;

			vS2.y+=-cp.y/cp.x;
			vT2.y+=-cp.z/cp.x;
			
			vS3.y+=-cp.y/cp.x;
			vT3.y+=-cp.z/cp.x;
		}

		// z, s, t
		edge01 = Vec3(v2.z-v1.z, Tex2.x-Tex1.x, Tex2.y-Tex1.y);
		edge02 = Vec3(v3.z-v1.z, Tex3.x-Tex1.x, Tex3.y-Tex1.y);
		cp = Cross(edge01, edge02);
		if(fabs(cp.x)>SMALL_FLOAT)
		{
			vS1.z+=-cp.y/cp.x;
			vT1.z+=-cp.z/cp.x;

			vS2.z+=-cp.y/cp.x;
			vT2.z+=-cp.z/cp.x;
			
			vS3.z+=-cp.y/cp.x;
			vT3.z+=-cp.z/cp.x;
		}
    }

    // Calculate the SxT vector
  	for(i=0; i<lNumVertices; i++)
	{
  		// Normalize the S, T vectors
		vS[i] = Normalize(vS[i]);
		vT[i] = Normalize(vT[i]);
  		
  		// Get the cross of the S and T vectors
		vBiNorm[i] = Cross(vS[i], vT[i]);
    		
  		// Get the direction of the SxT vector
  		if(vBiNorm[i]*Normals[i]<0.0f)
		{
  			vBiNorm[i] = -vBiNorm[i];
  		}

		vBiNorm[i] = -vT[i];
  	}

	int iTing;
	iTing = 0;
}

// Temporary stuff
#define CLUTADR			((((cScreen->GetWidth()*cScreen->GetHeight()*4*3)+(cScreen->GetWidth()*64*4))>>2)>>6)
#define CLUTADR1		((((cScreen->GetWidth()*cScreen->GetHeight()*4*3)+(cScreen->GetWidth()*64*4)+(256*4))>>2)>>6)

#define GS_PSMCT32			0x0
#define GS_PSMCT24			0x1
#define GS_PSMCT16			0x2

#define GS_PSMT8				0x13
#define GS_PSMT8H				0x1b

#define GS_PSMZ32				0x30
#define GS_PSMZ16S			0x3a

#define GS_PRIM_SPRITE		0x6




void StitchTstPage(CDmaVif1 * pChain)
{
	void * pCurLoc = pChain->GetCurLocation();
	uint32 uLowerDigits = ((uint32) pCurLoc)&0xfff;
	assert(uLowerDigits <= 0xff0);
	if(uLowerDigits == 0xff0)
	{
		pChain->EndPacket();
		
		// Insert a Next_Tag so DMAC 
		// can locate the next page
		pChain->AttachSrcNextTag( (void *) (((uint32) pCurLoc)+0x10) );
		pChain->EndPacket();

		// Relocate origional Tag to next page (but remove the VifCode)
		pChain->AttachSrcCntTag();
	}
}

void MakePacketAddRGB(CDmaVif1 * pChain,int frameFBP,int x,int y,int width,int height,int frameBufferWidth,int frameBufferHeight,int tempTextureFBP)
{
	int iVertStrip;

	// X, Y, WIDTH, HEIGHT in 12.4 FORMAT
	frameFBP>>=11;
	tempTextureFBP>>=11;
	if(height>256)asm("break");//INT3;

	width  = width<<4;
	height = height<<4;
	x = x<<4;
	y = y<<4;

	uint32 uQSizeDirPacket = (frameBufferWidth>>4)*4 + 16;

	StitchTstPage(pChain);
	pChain->add64(NOP<<32 | NOP);
	pChain->add64(DIRECT(uQSizeDirPacket)<<32 | FLUSH);

	uint32 uNumRegsToSet = uQSizeDirPacket - 1;

	StitchTstPage(pChain);
	pChain->add64(GIFTAG(uNumRegsToSet, 1, 0, 0, 0, 1));
	pChain->add64(GIFTAG_REGS1(AD));
	
	StitchTstPage(pChain);
	pChain->add64(XYOFFSET_SET(0,0));
	pChain->add64(XYOFFSET_1);

	StitchTstPage(pChain);
	pChain->add64(SCISSOR_SET(0, frameBufferWidth, 0, frameBufferHeight));
	pChain->add64(SCISSOR_1);

	StitchTstPage(pChain);
	pChain->add64(CLAMP_SET(0,0,0,0,0,0));
	pChain->add64(CLAMP_1);

	// This does the adding of Red, Green and Blue putting the results into the Alpha channel
	// The z-buffer might be a good place for a temp texture

	//----------------------------------------------------------------------------
	// 1st pass: Copy Red,Green into Blue, Alpha channel of 32bit temp texture (from newsgroup method)
	//----------------------------------------------------------------------------

	StitchTstPage(pChain);
	pChain->add64(FRAME_SET(tempTextureFBP,(frameBufferWidth>>6), GS_PSMCT16, 0));
	pChain->add64(FRAME_1);

	StitchTstPage(pChain);
	pChain->add64(ZBUF_SET(0, GS_PSMZ16S, 0));
	pChain->add64(ZBUF_1);

	StitchTstPage(pChain);
	pChain->add64(DTHE_SET(0));
	pChain->add64(DTHE);

	StitchTstPage(pChain);
	pChain->add64(TEST_SET(1,0,0,1,  0,0,1,1));									// Disable Z-tests and Z-writes FBwrite=RGBA
	pChain->add64(TEST_1);

	StitchTstPage(pChain);
	pChain->add64(ALPHA_SET(2, 2, 2, 0, 128));									// Disable alpha blend
	pChain->add64(ALPHA_1);

	StitchTstPage(pChain);
	pChain->add64(0);
	pChain->add64(TEXFLUSH);													// Flush texture cache out

	StitchTstPage(pChain);
	pChain->add64(TEX0_SET((frameFBP<<5),(frameBufferWidth>>6), GS_PSMCT16,  10, 10, 1, 1, 0, 0, 0, 0, 0));		// Texture source = FRAME
	pChain->add64(TEX0_1);

	StitchTstPage(pChain);
	pChain->add64(FBA_SET(0));
	pChain->add64(FBA_1);

	StitchTstPage(pChain);
	pChain->add64(TEXA_SET(0, 0, 0x80));
	pChain->add64(TEXA);

	StitchTstPage(pChain);
	pChain->add64(TEX1_SET(1, 0, 0, 0, 0, 0,0));								// Point sampling
	pChain->add64(TEX1_1);

	// Draw copy sprite to 0,0 in dest (half height)
	StitchTstPage(pChain);
	pChain->add64(PRIM_SET(GS_PRIM_SPRITE, 0, 1, 0, 1, 0, 1, 0, 0));
	pChain->add64(PRIM);

	StitchTstPage(pChain);
	pChain->add64(RGBAQ_SET(0x80, 0x80, 0x80, 0x80, 0x3f800000));
	pChain->add64(RGBAQ);


	for (iVertStrip=0; iVertStrip<(frameBufferWidth>>4); iVertStrip++)			// 8 pixel wide strips - skipping 8
	{
		StitchTstPage(pChain);
		pChain->add64(UV_SET(8+(iVertStrip<<8), 8));							// (+8,+8) for center of pixel
		pChain->add64(UV);

		StitchTstPage(pChain);
		pChain->add64(XYZ2_SET((8<<4)+(iVertStrip<<8), 0, 0));
		pChain->add64(XYZ2);

		StitchTstPage(pChain);
		pChain->add64(UV_SET(8+(iVertStrip<<8)+(8<<4), 8+(height<<1)));
		pChain->add64(UV);

		StitchTstPage(pChain);
		pChain->add64(XYZ2_SET((8<<4)+(iVertStrip<<8)+(8<<4), (height<<1), 0));
		pChain->add64(XYZ2);
	}



//----------------------------------------------------------------------------
// 2nd pass: Add tempTexture Blue, Alpha  to Blue, Alpha of frame buffer (alpha is replaced)
//----------------------------------------------------------------------------

	uQSizeDirPacket = (frameBufferWidth>>5)*4 + 9;
	StitchTstPage(pChain);
	pChain->add64(NOP<<32 | NOP);
	pChain->add64(DIRECT(uQSizeDirPacket)<<32 | FLUSH);

	uNumRegsToSet = uQSizeDirPacket - 1;
	StitchTstPage(pChain);
	pChain->add64(GIFTAG(uNumRegsToSet, 1, 0, 0, 0, 1));
	pChain->add64(GIFTAG_REGS1(AD));

	StitchTstPage(pChain);
	pChain->add64(0);
	pChain->add64(TEXFLUSH);													// Flush texture cache out

	StitchTstPage(pChain);
	pChain->add64(FRAME_SET(tempTextureFBP, (frameBufferWidth>>6), GS_PSMCT32, 0xff00ffff));
	pChain->add64(FRAME_1);

	StitchTstPage(pChain);
	pChain->add64(ZBUF_SET(0, GS_PSMZ32, 0));
	pChain->add64(ZBUF_1);

	StitchTstPage(pChain);
	pChain->add64(TEST_SET(1,0,0,1,  0,0,1,1));									// Disable Z-tests and Z-writes FBwrite=RGBA
	pChain->add64(TEST_1);

	StitchTstPage(pChain);
	pChain->add64(TEX0_SET((frameFBP<<5),(frameBufferWidth>>6), GS_PSMCT32, 10, 10, 1, 1, 0, 0, 0, 0, 0)); // Texture source = tempTextureFBP
	pChain->add64(TEX0_1);

	StitchTstPage(pChain);
	pChain->add64(ALPHA_SET(0, 2, 2, 1, 128)); // Set alpha blend to add
	pChain->add64(ALPHA_1);

	StitchTstPage(pChain);
	pChain->add64(PRIM_SET(GS_PRIM_SPRITE, 0, 1, 0, 1, 0, 1, 0, 0));
	pChain->add64(PRIM);

	StitchTstPage(pChain);
	pChain->add64(RGBAQ_SET(0x80,0x80,0x80,0x80,0x3f800000));
	pChain->add64(RGBAQ);

	for (iVertStrip=0; iVertStrip<(frameBufferWidth>>5); iVertStrip++) // 32 pixel wide strips
	{
		StitchTstPage(pChain);
		pChain->add64(UV_SET(8+(iVertStrip<<9), 8));
		pChain->add64(UV);
		
		StitchTstPage(pChain);
		pChain->add64(XYZ2_SET((iVertStrip<<9), 0, 0));
		pChain->add64(XYZ2);

		StitchTstPage(pChain);
		pChain->add64(UV_SET(8+(32<<4)+(iVertStrip<<9), 8+height));
		pChain->add64(UV);

		StitchTstPage(pChain);
		pChain->add64(XYZ2_SET((32<<4)+(iVertStrip<<9), height, 0));
		pChain->add64(XYZ2);
	}


//----------------------------------------------------------------------------
// 3rd pass: Add Alpha to Blue using a CLUT with PSMCT8H mode ( IN-PLACE )
//----------------------------------------------------------------------------
// Use PSMCT24 mode to mask write to Alpha channel

	uQSizeDirPacket = (frameBufferWidth>>5)*4 + 9;
	StitchTstPage(pChain);
	pChain->add64(NOP<<32 | NOP);
	pChain->add64(DIRECT(uQSizeDirPacket)<<32 | FLUSH);

	uNumRegsToSet = uQSizeDirPacket - 1;
	StitchTstPage(pChain);
	pChain->add64(GIFTAG(uNumRegsToSet, 1, 0, 0, 0, 1));
	pChain->add64(GIFTAG_REGS1(AD));

	StitchTstPage(pChain);
	pChain->add64(0);
	pChain->add64(TEXFLUSH);

	StitchTstPage(pChain);
	pChain->add64(FRAME_SET(tempTextureFBP,(frameBufferWidth>>6), GS_PSMCT24, 0xff00ffff));
	pChain->add64(FRAME_1);

	StitchTstPage(pChain);
	pChain->add64(ZBUF_SET( 0, GS_PSMZ32, 0));
	pChain->add64(ZBUF_1);

	StitchTstPage(pChain);
	pChain->add64(TEST_SET(1,0,0,1,  0,0,1,1)); // Disable Z-tests and Z-writes FBwrite=RGBA
	pChain->add64(TEST_1);

	// Use texture clut (0-255) in Blue channel (or in EVERY channel)
	StitchTstPage(pChain);
	pChain->add64(TEX0_SET((tempTextureFBP<<5),(frameBufferWidth>>6), GS_PSMT8H,10,10, 1, 0, CLUTADR, 0, 0,	0, 1)); // TextureSrc =frame buffer (PSMCT8H mode) CLD=1(Load clut)
	pChain->add64(TEX0_1);

	StitchTstPage(pChain);
	pChain->add64(ALPHA_SET(1, 2, 2, 0, 128));   // Set alpha blend to add (ALREADY DONE)
	pChain->add64(ALPHA_1);

	StitchTstPage(pChain);
	pChain->add64(PRIM_SET(GS_PRIM_SPRITE, 0, 1, 0, 1, 0, 1, 0, 0));
	pChain->add64(PRIM);

	// Draw sprite to screen
	StitchTstPage(pChain);
	pChain->add64(RGBAQ_SET(0x80,0x80,0x80,0x80,0x3f800000));
	pChain->add64(RGBAQ);


	for (iVertStrip=0; iVertStrip<(frameBufferWidth>>5); iVertStrip++) // 32 pixel wide strips
	{
		StitchTstPage(pChain);
		pChain->add64(UV_SET(8+(iVertStrip<<9), 8));
		pChain->add64(UV);
		
		StitchTstPage(pChain);
		pChain->add64(XYZ2_SET((iVertStrip<<9), 0, 0));
		pChain->add64(XYZ2);

		StitchTstPage(pChain);
		pChain->add64(UV_SET(8+(32<<4)+(iVertStrip<<9),8+height));
		pChain->add64(UV);

		StitchTstPage(pChain);
		pChain->add64(XYZ2_SET((32<<4)+(iVertStrip<<9),height, 0));
		pChain->add64(XYZ2);
	}



//----------------------------------------------------------------------------
// 4th pass: Move Blue into the Alpha channel
//----------------------------------------------------------------------------

// Note: Each 32bit page is 64x32, each 8bit page is 128x64

	for(uint32 a=0;a!=20;a++)
	{
		uint32 pageFBP=frameFBP+a;
		uint32 pageFBP1=tempTextureFBP+a;

		uQSizeDirPacket = 8 + 2*8*(8+4);
		StitchTstPage(pChain);
		pChain->add64(NOP<<32 | NOP);
		pChain->add64(DIRECT(uQSizeDirPacket)<<32 | FLUSH);

		uNumRegsToSet = 6;
		StitchTstPage(pChain);
		pChain->add64(GIFTAG(uNumRegsToSet, 1, 0, 0, 0, 1));
		pChain->add64(GIFTAG_REGS1(AD));

		StitchTstPage(pChain);
		pChain->add64(FRAME_SET(pageFBP,(frameBufferWidth>>6), GS_PSMCT32, 0x00ffffff));
		pChain->add64(FRAME_1);

		StitchTstPage(pChain);
		pChain->add64(TEX0_SET((pageFBP1<<5),2,GS_PSMT8,10, 10, 1, 1, CLUTADR1, 0, 0, 0, 1 ));
		pChain->add64(TEX0_1);

		StitchTstPage(pChain);
		pChain->add64(CLAMP_SET(3,0,0xf7,0x8,0,0));
		pChain->add64(CLAMP_1);

		StitchTstPage(pChain);
		pChain->add64(ALPHA_SET(2, 2, 2, 0, 128));
		pChain->add64(ALPHA_1);

		StitchTstPage(pChain);
		pChain->add64(PRIM_SET(GS_PRIM_SPRITE, 0, 1, 0, 1, 0, 1, 0, 0));
		pChain->add64(PRIM);

		StitchTstPage(pChain);
		pChain->add64(RGBAQ_SET(0x80,0x80,0x80,0x80,0x3f800000));
		pChain->add64(RGBAQ);

		// add strips
		uNumRegsToSet = 8*(8+4);
		StitchTstPage(pChain);
		pChain->add64(GIFTAG(uNumRegsToSet, 1, 0, 0, 1, 4));	// reglist
		pChain->add64(GIFTAG_REGS4(UV, XYZ2, UV, XYZ2));
		

		uint32 lCount=0;
		for (y=0; y<32; y+=2)
		{
			if ((y%4)==0)
			{
				// EVEN COLUMN (Four 16x2 pixel sprites)
				for (x=0; x<64; x+=16) // 4 sprites
				{
					// Using region repeat mask to skip 8 texel so you can use a 16 wide sprite
					// XY [0,0 to 16,2]
					StitchTstPage(pChain);
					pChain->add64(UV_SET(8 + ((8+x*2)<<4) , 8 + ((y*2)<<4)));
					pChain->add64(XYZ2_SET( (x<<4)        , (y<<4), 0 ));
					StitchTstPage(pChain);
					pChain->add64(UV_SET(8 + ((24+x*2)<<4), 8 + ((2+y*2)<<4) ));
					pChain->add64(XYZ2_SET( ((x+16)<<4)    , ((y+2)<<4), 0 ));

					++lCount;
				}
			}
			else
			{
				// ODD COLUMN (Eight 8x2 pixel sprite)
				for (x=0; x<64; x+=8) // 8 sprites
				{
					// Using region repeat mask to swap 4x2 texture regions over while fetching texels
					// XY [0,0 to 8,2]
					StitchTstPage(pChain);
					pChain->add64(UV_SET(8 + ((4+x*2)<<4) , 8 + ((y*2)<<4) ));
					pChain->add64(XYZ2_SET( (x<<4)        , (y<<4), 0 ));
					StitchTstPage(pChain);
					pChain->add64(UV_SET(8 + ((12+x*2)<<4), 8 + ((2+y*2)<<4) ));
					pChain->add64(XYZ2_SET( ((x+8)<<4)    , ((y+2)<<4), 0 ));

					++lCount;
				}
			}
		}

		assert(lCount == 8*(8+4));
	}
}




void InitRegs()
{
	CDmaVif1 VifChain(pAllDmaMem,16);
	VifChain.AttachSrcEndTag();
	VifChain.add64(NOP<<32 | NOP);
	VifChain.add64(DIRECT(9)<<32 | FLUSH);

	VifChain.add64(GIFTAG(8, 1, 0, 0, 0, 1));
	VifChain.add64(GIFTAG_REGS1(AD));

	VifChain.add64(0);
	VifChain.add64(TEXFLUSH);

	VifChain.add64(TEX1_SET(0, 0, 0, 0, 0, 0,0));
	VifChain.add64(TEX1_1);
	
	VifChain.add64(TEST_SET(0, 0, 0, 0, 0, 0, 0, 1));
	VifChain.add64(TEST_1);

	VifChain.add64(CLAMP_SET(0, 0, 0, 0, 0, 0));
	VifChain.add64(CLAMP_1);

	VifChain.add64(ALPHA_SET(0, 0, 0, 0, 0));
	VifChain.add64(ALPHA_1);

	VifChain.add64(TEXA_SET(0, 0, 0xffl));
	VifChain.add64(TEXA);

	VifChain.add64(ZBUF_SET(0, 0, 1));
	VifChain.add64(ZBUF_1);

	VifChain.add64(FOGCOL_SET(0, 0, 0));
	VifChain.add64(FOGCOL);

	VifChain.EndPacket();
	VifChain.Launch(iSPS2Device);
	VifChain.Wait();
}


void MakePacketAddRGB( CDmaVif1 * pChain,int frameFBP, int x, int y, int width, int height,int frameBufferWidth,int frameBufferHeight, int tempTextureFBP);

void SendCluttoVRAM(void * data, uint32 texAddr, uint32 clutAddr)
{
	CDmaVif1 dmaVif1(pAllDmaMem,8);
	
	dmaVif1.AttachSrcCntTag(DIRECT(6)<<32);

	dmaVif1.add64(GIFTAG(4, 0, 0, 0, 0, 1));
	dmaVif1.add64(GIFTAG_REGS1(AD));

	dmaVif1.add64(BITBLTBUF_SET(0, 0, 0, clutAddr, 1, 0));		// 0x41a40 00000000L
	dmaVif1.add64(BITBLTBUF);
	dmaVif1.add64(TRXPOS_SET(0, 0, 0, 0, 0));
	dmaVif1.add64(TRXPOS);
	dmaVif1.add64(TRXREG_SET(16, 16));
	dmaVif1.add64(TRXREG);
	dmaVif1.add64(TRXDIR_SET(0));
	dmaVif1.add64(TRXDIR);
	dmaVif1.add64(GIFTAG(1024/16, 1, 0, 0, 2, 0));	// GIFTAG(NLOOP, EOP, PRE, PRIM, FLG, NREG)
	dmaVif1.add64(0);
	dmaVif1.EndPacket();
	dmaVif1.AttachSrcRefeTag(data, 1024/16, DIRECT(1024/16)<<32);

	dmaVif1.Launch(iSPS2Device);
	dmaVif1.Wait();

	//GsSyncPath(0,0);
}

void DownLoadPals()
{
	// This is for linear CLUT (used for add blue to alpha)
	int i;
	
	uint32 * myPal = (uint32 *) PaletteUnc;

	for (i=0; i<256; i++)
	{
		myPal[(i&0xe7) | ((i&0x08)<<1) | ((i&0x10)>>1)] = i + (i<<8) + (i<<16) + (i<<24);
	}
	SendCluttoVRAM(PaletteCach, 0, CLUTADR);

	// This is clamped CLUT
	int ambient = 15;
	for (i=0; i<256; i++)
	{
		if (i>128) // clamp negative values to 0 (ambient level)
		{       
			myPal[(i&0xe7) | ((i&0x08)<<1) | ((i&0x10)>>1)]=ambient+(ambient<<8)+(ambient<<16)+(ambient<<24);
		}else{ // positive values have ambient component added
			/*int red  = (i<<1) + ambient;
			int green= (i<<1) + ambient;
			int blue = (i<<1) + ambient;
			int alpha= (i<<1) + ambient;
			if(red>255)red=255;
			if(green>255)green=255;
			if(blue>255)blue=255;
			if(alpha>255)alpha=255;
			*/
			int red  = (i) + ambient;
			int green= (i) + ambient;
			int blue = (i) + ambient;
			int alpha= (i) + ambient;
			if(red>255)red=255;
			if(green>255)green=255;
			if(blue>255)blue=255;
			if(alpha>255)alpha=255;

			myPal[(i&0xe7) | ((i&0x08)<<1) | ((i&0x10)>>1)] = red + (green<<8) + (blue<<16) + (alpha<<24);
		}
	}
	SendCluttoVRAM(PaletteCach, 0, CLUTADR1);
}


CDmaVif1 BuildAddRGBintoAlpha(uint32 uDmaMem, uint32 lFrame,uint32 lTemp)
{
	InitRegs();

	CDmaVif1 VifChain((void *) uDmaMem, (1024*1024)/16);
	VifChain.DisableTTE();
	VifChain.DisableUCAB();

	uint32 uStop = (cScreen->GetMode() == SCR_PAL) ? 8 : 7;
	for(uint32 a=0;a!=uStop;a++)
	{
		StitchTstPage(&VifChain);
		VifChain.AttachSrcCntTag();
		MakePacketAddRGB(&VifChain,(lFrame<<11)+(a*((cScreen->GetWidth()*64*4)>>2)),0,0,cScreen->GetWidth(),64,cScreen->GetWidth(),cScreen->GetHeight(),lTemp<<11);
		VifChain.EndPacket();
	}

	// if Vesa
	if(cScreen->GetMode() == SCR_VESA)
	{
		StitchTstPage(&VifChain);
		VifChain.AttachSrcCntTag();
		MakePacketAddRGB(&VifChain,(lFrame<<11)+(7*((cScreen->GetWidth()*64*4)>>2)),0,0,cScreen->GetWidth(),32,cScreen->GetWidth(),cScreen->GetHeight(),lTemp<<11);
		VifChain.EndPacket();
	}

	// Finish up
	StitchTstPage(&VifChain);
	VifChain.AttachSrcEndTag();
	VifChain.EndPacket();
	
	return VifChain;
}
