#include "camera.h"

void CCamera::SetClipPlanes(float fFrontClip, float fBackClip)	// Should both be negative
{
	m_fFrontClip = fFrontClip;
	m_fBackClip = fBackClip;

	SetupMatrices();
}

void CCamera::SetPosViewPlane(const Vec3 &vVRP)
{
	m_vVRP = vVRP;

	SetupMatrices();
}

void CCamera::SetRotViewPlane(const Vec3 &vVPN, const Vec3 &vVUP)
{
	m_vVPN = Normalize(vVPN);
	m_vVUP = Normalize(vVUP);

	SetupMatrices();
}

void CCamera::SetBoundVRC(float fUmin, float fVmin, float fUmax, float fVmax)
{
	m_fUmin=fUmin; m_fVmin=fVmin;
	m_fUmax=fUmax; m_fVmax=fVmax;

	SetupMatrices();
}

void CCamera::SetEyePos(const Vec3 &vPRP)
{
	m_vPRP = vPRP;

	SetupMatrices();
}

void CCamera::SetBitmapSize(float fPixelWidth, float fPixelHeight, float fZBufDepth)
{
	m_fPixelWidth = fPixelWidth;
	m_fPixelHeight = fPixelHeight;
	m_fZBufDepth = fZBufDepth;

	SetupMatrices();
}

// if viewport is located at 0,0 (lower left)
// then leave these values at 0 aswell
void CCamera::SetViewPortOffset(float fViewOffsX, float fViewOffsY, float fViewOffsZ)
{
	m_fViewPortOffsX = fViewOffsX;
	m_fViewPortOffsY = fViewOffsY;
	m_fViewPortOffsZ = fViewOffsZ;

	SetupMatrices();
}

void CCamera::SetupMatrices()
{
	// Get World Transform
	Vec3 Uaxis = Normalize(Cross(m_vVUP, m_vVPN));
	Vec3 Vaxis = Normalize(Cross(m_vVPN, Uaxis));	// Normalize just in case

	SetColumn(&m_m44TransformCamToWorld, 0, Uaxis);
	SetColumn(&m_m44TransformCamToWorld, 1, Vaxis);
	SetColumn(&m_m44TransformCamToWorld, 2, m_vVPN);
	SetColumn(&m_m44TransformCamToWorld, 3, m_vVRP);
	SetRow(&m_m44TransformCamToWorld, 3, Vec4(0,0,0,1));
	
	// Get Perspective Transform
	Mat44 m44M, m44Sper, m44SHpar, m44Tprp;

	LoadIdentity(&m44M);
	LoadIdentity(&m44Sper);
	LoadIdentity(&m44SHpar);
	LoadIdentity(&m44Tprp);

	// Get TRP
	SetColumn(&m44Tprp, 3, (Vec4) -m_vPRP);

	// Get SH
	Vec3 vCW((m_fUmax+m_fUmin)*0.5f, (m_fVmax+m_fVmin)*0.5f, 0);
	Vec3 vDOP = vCW - m_vPRP;

	SetColumn(&m44SHpar, 2, Vec4(-vDOP.x/vDOP.z, -vDOP.y/vDOP.z, 1, 0) );

	// Get S
	Vec3 vrp = (Vec3) (m44SHpar * (m44Tprp * Vec4(0,0,0,1)));

	float fSx = (2 * vrp.z) / ( (m_fUmax-m_fUmin)*(vrp.z+m_fBackClip) );
	float fSy = (2 * vrp.z) / ( (m_fVmax-m_fVmin)*(vrp.z+m_fBackClip) );
	float fSz = -1 / (vrp.z+m_fBackClip);

	SetColumn(&m44Sper, 0, Vec4(fSx, 0, 0, 0));
	SetColumn(&m44Sper, 1, Vec4(0, fSy, 0, 0));
	SetColumn(&m44Sper, 2, Vec4(0, 0, fSz, 0));

	// Get M
	float fZmin = -(vrp.z+m_fFrontClip)/(vrp.z+m_fBackClip);

	SetColumn(&m44M, 2, Vec4(0, 0, 1 / (1 + fZmin), -1) );
	SetColumn(&m44M, 3, Vec4(0, 0, -fZmin / (1 + fZmin), 0) );

	// Sum up
	m_m44PerspectiveTransform = m44M * m44Sper * m44SHpar * m44Tprp;


	// Get Viewport Transform
	Mat44 Sview, Tview;

	SetColumn(&Sview, 0, Vec4(m_fPixelWidth*0.5f, 0, 0, 0) );
	SetColumn(&Sview, 1, Vec4(0, m_fPixelHeight*0.5f, 0, 0) );
	SetColumn(&Sview, 2, Vec4(0, 0, m_fZBufDepth, 0) );
	SetColumn(&Sview, 3, Vec4(0, 0, 0, 1) );

	LoadIdentity(&Tview);
	SetColumn(&Tview, 3, Vec4(m_fViewPortOffsX, m_fViewPortOffsY, m_fViewPortOffsZ, 1) );

	m_m44ViewportTransform = Tview * Sview;

    // RoundUp all
	m_m44WorldToBitmapTransform = m_m44ViewportTransform * m_m44PerspectiveTransform * (~m_m44TransformCamToWorld);
}

const Mat44 CCamera::GetWorldToBitmapTransform()		// Alle Matrices put together
{
	return m_m44WorldToBitmapTransform;
}

const Mat44 CCamera::GetTransformCamToWorld()
{
	return m_m44TransformCamToWorld;
}

const Mat44 CCamera::GetPerspectiveTransform()
{
	return m_m44PerspectiveTransform;
}

const Mat44 CCamera::GetViewportTransform()
{
	return m_m44ViewportTransform;
}


CCamera::CCamera()
{
	m_vPRP = Vec3(0,0,10);
	m_fUmin=-17, m_fVmin=-13;
	m_fUmax=17, m_fVmax=13;

	m_vVRP = Vec3(0,0,0); 
	m_vVPN = Vec3(0,0,1); 
	m_vVUP = Vec3(0,1,0);

	m_fFrontClip = -50; 
	m_fBackClip = -10000;

	m_fPixelWidth = 512;
	m_fPixelHeight = 512;
	m_fZBufDepth = 1;

	m_fViewPortOffsX = 0;
	m_fViewPortOffsY = 0;

	SetupMatrices();
}
